!************************************************
! Test program for omdetect- checks the output
!************************************************
!  VNY 2.11.2001 division operations are made more robust
program checkoutput
use dal
use errorhandling

implicit none

    type(DataSetT)  :: listDataSet
    type(TableT)    :: tableDataSet
    type(ColumnT)   :: listColumn1,listColumn2,listColumn3,listColumn4
    type(ColumnT)   :: listColumn5, listColumn6, listColumn7, listColumn8
    real(kind=single), pointer :: xpos(:),ypos(:),poserr(:),counts(:),cerr(:)
    real(kind=single), pointer :: sigX(:), sigY(:),pa(:)
    real, allocatable :: xSource(:), ySource(:), cSource(:)
    real, allocatable :: sigXSource(:), sigYSource(:), paSource(:)
    integer, dimension(:), allocatable :: index
    integer :: nSource
    integer :: nFound
    real :: chiSqr
    integer :: i, j, k, l, nrows
    real :: dist
    logical, allocatable :: Found(:)
    integer :: thisSource
    real :: mindist
    real :: radius = 6.0 ! Was 25 - okay with 6
    character(len=32) :: xpout
    character(len=32) :: ypout
    integer noFiles
    integer :: jStart, Locate
    external Locate
    character(len=50) :: fileName
    character(len=100) :: text    
    real, dimension(2) :: passNo = (/0.70, 0.3/)

    !****************************************************
    ! The progam createtest.f90 will create one or more
    ! test images (and lists of the sources that it has 
    ! created (testlistn.lis". 
    ! Omdetect will then produce, for each image,
    ! a source-list file called "testlistn.fts". 
    ! n is 1,2, etc, for the number of tests created by
    ! createtest.f90.
    ! Currently doing two tests- one for omichain images
    ! and one for omfchain images.
    !************************************************** 
    noFiles = 2
            
    do l = 1, noFiles
       
        write(fileName, '(a, i1, a)') "testlist", l, ".lis"
             
        open(20, file = trim(fileName) , status = "old", form = "formatted")
   
        write(*, *) "Checking output from testlist ", trim(filename), " produced by omdetect"
   
        write(fileName, '(a, i1, a)') "testlist", l, ".fts"
 
        read(20, *) nSource ! The number of sources that were created for the image
        listDataSet = dataSet(trim(fileName), read)
    
        tableDataSet = table(listDataSet, "SRCLIST")
        nrows = numberofRows(tableDataSet)

        write(*, '(a, i4)') "No of sources in test image=", nSource
        if (nrows .eq. 0) call fatal("No sources detected", text)

        listColumn1 = column(tableDataSet, "XPOS",read)

        xpos => real32Data(listColumn1)
        listColumn2 = column(tableDataSet, "YPOS",read)
        ypos => real32Data(listColumn2)
        listColumn3 = column(tableDataSet, "POSERR",read)
        poserr => real32Data(listColumn3)
        listColumn4 = column(tableDataSet, "RATE",read)
        counts => real32Data(listColumn4)
        listColumn5 = column(tableDataSet, "RATE_ERR",read)
        cerr => real32Data(listColumn5)
        listColumn6 = column(tableDataSet, "FWHM_MAJ",read)
        sigX => real32Data(listColumn6)
        listColumn7 = column(tableDataSet, "FWHM_MIN",read)
        sigY => real32Data(listColumn7)
        listColumn8 = column(tableDataSet, "PA",read)
        pa => real32Data(listColumn8)


        !********************************************
        ! Sort this data into increasing x order
        !*******************************************

        call SortData(nrows, xpos, ypos, poserr, counts, cerr, sigX, sigY, pa)
                
        write(*, *) "No sources created for test=", nsource
        if (nrows .gt. nsource) then
            write(text, '(a, i4, a, i4)') "No of sources detected =", nrows, " actual no present=", nsource
       
            call fatal("Wrong no of sources detected", text)
        end if
        allocate(xSource(0 : nSource - 1))
        allocate(ySource(0 : nSource - 1))
        allocate(cSource(0 : nSource - 1))
        allocate(sigXSource(0 : nSource - 1))
        allocate(sigYSource(0 : nSource - 1))
        allocate(paSource(0 : nSource - 1))

        allocate(Found(0 : nSource - 1))
        allocate(index(0 : nSource - 1))

        Found = .false.

       !**********************************
       ! Read in the source list
       !**********************************

        do i = 0, nSource - 1
            read(20, *) xSource(i), ySource(i), cSource(i), sigXsource(i), &
                   sigYsource(i), paSource(i)
        end do

        !***********************************************
        ! Sort the sources into increasing x-order
        !***********************************************
        call IndexSort(xSource, index, nSource)


        nFound = 0

        chiSqr = 0.0
        thissource = 0
        
        jStart = 0
        do i = 0, nSource - 1
            k = index(i)

            mindist = 1.0e30
!            jStart = Locate(nRows, xpos, xSource(k), jStart)
!            jStart = max(0, jStart - 20)
            do j = jStart, nrows - 1
                if (found(j)) cycle
                write (xpout, *) xpos(j)
                write (ypout, *) ypos(j)                                     
                if (scan(xpout,"NaN") .eq. 0 .and. scan(ypout,"NaN") .eq. 0 ) then
                    dist = (xSource(k) - xpos(j)) ** 2 + (ySource(k) - ypos(j)) ** 2
                    dist = sqrt(dist)
                    if (dist .lt. mindist) then
                        mindist = dist
                        thissource = j
                    end if
                end if
            end do

            mindist = sqrt(mindist)
            if (mindist .lt. radius * poserr(thissource)) then
                nFound = nFound + 1
!---VNY 2.11.2001 More robust division
                if (sqrt(cSource(k)) .gt. 1.0e-10) then
                   chiSqr = chiSqr + ((cSource(k) - counts(thissource)) / sqrt(cSource(k))) ** 2
                end if
                write(*, 200) "Source no ", k, " found (", xSource(k), ",", ySource(k), &
                        ") error=", poserr(thissource),  "  omdetect position=(", xpos(thissource), ",", ypos(thissource), &
                        ") minimum distance=", mindist
                found(thissource) =  .true.         
 
            else
                write(*, 200) "Source no ", k, " not found (", xSource(k), ",", ySource(k), ")" 
   
            end if
            
        end do


200 format(a, i6, a, f8.2, a, f8.2, a, f6.2, a, f8.2, a, f8.2, a, f6.2, a)
       

!---VNY 2.11.2001 More robust division
        if ((nFound-1) .gt. 0) then
           chiSqr = chiSqr / real(nFound-1)
        end if
        write(*, *) "ChiSqr of flux values: ",chiSqr
        write(*, *) "Number of sources in test image: ", nSource
        write(*, *) "Number of sources found by omdetect: ", nFound

        if (nFound .gt. nSource+2 .or. abs(real(nFound-nSource)/real(nSource)) .gt. passNo(l)) then  ! NB was 0.3
            write(text, '(a, i4, a, i4)') "No of sources detected =", nrows, " actual no present=", nsource
          
            call fatal("wrongNumberOfSources", text)
        end if
    
        deallocate(xSource)
        deallocate(ySource)
        deallocate(cSource)
        deallocate(sigXSource)
        deallocate(sigYSource)
        deallocate(paSource)
        deallocate(Found)
        deallocate(index)

        call release(listColumn1)
        call release(listColumn2)
        call release(listColumn3)
        call release(listColumn4)
        call release(listColumn5)
        call release(listColumn6)
        call release(listColumn7)
        call release(listColumn8)
        call release(tableDataSet)
        call release(listDataSet)
    end do
    close(20)
end


!********************************************
! Subroutine SortData
!*******************************************
subroutine SortData(nRows, xpos, ypos, poserr, counts, cerr, sigX, sigY, pa)
use dal    
    integer, intent(in) :: nRows
    real(kind=single), dimension(0 : nRows - 1) :: xpos, ypos, poserr, &
                                       counts, cerr, sigX, sigY, pa
    real(kind=single), dimension(:), allocatable :: tempArray
    integer, dimension(:), allocatable :: index
    integer i, size
    size = nRows - 1
    allocate(index(0 : size))



    call IndexSort(xpos, index, nRows)

    allocate(tempArray(0 : size))
    
    tempArray = xpos
    forall(i = 0 : size) xpos(i) = tempArray(index(i))

    tempArray = ypos
    forall(i = 0 : size) ypos(i) = tempArray(index(i))
     
    tempArray = posErr
    forall(i = 0 : size) posErr(i) = tempArray(index(i))


    tempArray = counts
    forall(i = 0 : size) counts(i) = tempArray(index(i))

    tempArray = cerr
    forall(i = 0 : size) cerr(i) = tempArray(index(i))

    tempArray = sigX
    forall(i = 0 : size) sigX(i) = tempArray(index(i))


    tempArray = sigY
    forall(i = 0 : size) sigY(i) = tempArray(index(i))


    tempArray = pa
    forall(i = 0 : size) pa(i) = tempArray(index(i))

    deallocate(tempArray)
    deallocate(index)

end subroutine SortData

!************************************************************
! Subroutine HeapSort
! Sorts  one-dimensional array(1.. size) int ascending order.
! An index of the sorted order relative to the original
! order is returned.
!******************** 

subroutine IndexSort(array1, index, size)
    use dal
    integer, intent(in) :: size
    real(kind=single), intent(in), dimension(0 : size - 1) :: array1
    integer, intent(out), dimension(0 : size - 1) :: index

    integer :: i, ir, j, l, irra
    real(kind=double) :: rra

    do i = 0, size - 1
        index(i) = i
    end do   

return
    if (size < 2) return


    l = size / 2 + 1
    ir = size   
  
    do 
        if (l > 1) then
            l = l - 1
            rra = array1(index(l - 1))   
            irra = index(l - 1)    
        else
            rra = array1(index(ir - 1))        
            irra = index(ir - 1)
          
            index(ir - 1) = index(0)
            ir = ir - 1
            
            if (ir <= 1) then
             
                index(0) = irra
                exit
            end if
        end if

        i = l
        j = l + l
        do while(j <= ir)
       
            if (j < ir) then
                if(array1(index(j - 1)) < array1(index(j))) j = j + 1
            end if
            if (rra < array1(index(j - 1))) then
               
                index(i - 1) = index(j - 1)
                i = j
                j = j + j
            else
                exit
            end if
            
        end do   
          
        index(i - 1) = irra
    end do
  
end subroutine IndexSort


!********************************************
! Function Locate
! Given an array table(1 .. n), and
! given a value x, returns the index such that 
! x is between table(j) and table(j+1)
! The search starts at the position jStart
!********************************************
integer function Locate(size, table1, searchValue, jStart)
    use dal
    implicit none
    integer, intent(in) :: size
    real(kind=single), intent(in), dimension(0 : size - 1) :: table1
    real(kind=single), intent(in) :: searchValue
    integer, intent(inout) :: jStart
    integer :: k, klo, khi

    !********************************************
    ! Check to see if x is less than the minimum
    ! value in the table
    !********************************************
    if (searchValue .lt. table1(0)) then
        Locate = 0
        jStart = 0
        return
    end if
  
    !***************************************
    ! Check to see if x is greater than the 
    ! maximum value in the table
    !***************************************
    if (searchValue .gt. table1(size - 1)) then
        Locate = size - 1
        return
    end if

    klo = jStart + 1
    khi = size

    do while((khi - klo) .gt. 1)
        k = (khi + klo) / 2
        if (table1(k - 1) .gt. searchValue) then
            khi = k
        else
            klo = k
        end if
    end do

    !*********************************
    ! klo and khi now bracket the 
    ! searchvalue
    !*********************************
    Locate = klo - 1
    jStart = klo - 1
end function Locate 


