!******************************************
! Test program for omdetect
!******************************************
!---02.11.2001 VNY more robust calculation of exp() was made in order to 
!              avoid underflow messages
program CreateTest

use dal

implicit none
    !***************************************
    ! Create an image for the fast-imaging chain
    !**************************************
    call CreateTestImage(24, 24, 3, "testimage1.fts", "testlist1.lis", "FAST")

    !***************************************
    ! Create an image for the imaging chain
    !**************************************
    call CreateTestImage(256, 256, 300, "testimage2.fts", "testlist2.lis", "IMAGING")

end program createtest
!****************************************
! Subroutine CreateTestImage
!****************************************
Subroutine CreateTestImage(nx, ny, nSource, imageFileName, srcListFileName, dataMode)
use dal
    integer, intent(in) :: nx, ny, nSource
    character(len=*), intent(in) :: imageFileName, srcListFileName
    character(len=*), intent(in) :: dataMode
    real r1

    integer :: dims(2)
    integer :: i,j,k
    real, parameter  :: fracExtend=0.02, fracPoint = (1. - fracExtend)
    real, parameter :: exposure = 1000.
    real :: r, delta, xDash, yDash, ex1, ex2, F
    real :: background
    real :: flux(nSource)
    real :: sigX(nSource)
    real :: sigY(nSource)
    real :: sigXTimesSigY(nSource)
    real :: pa(nSource)
    real :: xSource(nSource)
    real :: ySource(nSource)
    real :: rSourceFlux(nSource)
    real, parameter :: pi = 3.141592654
    type(DataSetT) :: outSet
    type(ArrayT) :: outArray 
    real(kind=single), pointer :: image(:,:)
    real :: minFlux, maxFlux
    real, parameter :: nSigma = 10
    real, parameter :: maxSig = 1., minSig=0.25
    real :: x, y
    real :: tot
    real :: gasdev
    real :: meanval


!
!-----------------------------------------------------------------
!
! AMK
!
    type(TableT) :: modesTableH                        ! Modes table dataset handle.
    type(ColumnT), dimension(5) :: modesColumnH        ! Modes column handle.
    integer(kind=int32), dimension(:), pointer :: modeP! Mode pointer.
    integer(kind=int32), dimension(:), pointer :: windowX0P           ! WindowX0 pointer.
    integer(kind=int32), dimension(:), pointer :: windowY0P           ! WindowY0 pointer.
    integer(kind=int32), dimension(:), pointer :: windowDXP           ! WindowDX pointer.
    integer(kind=int32), dimension(:), pointer :: windowDYP           ! WindowDY pointer.
    integer(kind=int32), dimension(1) :: columnDim                    ! Column dimension.

    type(ArrayT) :: qualArrayH                                        ! Quality Array handle.
    integer(kind=int32), dimension(2) :: qualArrayDim                 ! Quality array dimension.

    !---VNY 2.11.2001
    integer(kind=int8), pointer :: qualArrayImage(:,:)
!
!-----------------------------------------------------------------
!

    dims(1) = nx
    dims(2) = ny

    call random_seed()

    outSet = dataSet(imageFileName,create)
    outArray = addArray(outSet, "PRIMARY", real32, dims)
    image => real32array2Data(outArray)

!
!------------------------------------------------------------------------
!
! AMK
!
! THis just to keep OMDETECT happy !!!!
!
! Add a MODES extension, detailing the science window parameters.
!
     
    modesTableH = addTable(outset, "MODES", 1)
    columnDim(1) = 1

    modesColumnH(1) = addColumn(modesTableH, "MODE", Integer32, "none",  columnDim)
    modeP => int32Data(modesColumnH(1))
    modeP = 0

    modesColumnH(2) = addColumn(modesTableH, "WINDOWX0", Integer32, "pixels",  columnDim)
    windowX0P => int32Data(modesColumnH(2))
    windowX0P = 0

    modesColumnH(3) = addColumn(modesTableH, "WINDOWY0", Integer32, "pixels",  columnDim)
    windowY0P => int32Data(modesColumnH(3))
    windowY0P = 0

    modesColumnH(4) = addColumn(modesTableH, "WINDOWDX", Integer32, "pixels",  columnDim)
    windowDXP => int32Data(modesColumnH(4))
    windowDXP = 0

    modesColumnH(5) = addColumn(modesTableH, "WINDOWDY", Integer32, "pixels",  columnDim)
    windowDYP => int32Data(modesColumnH(5))
    windowDYP = 0

    ! Release handles on data.
    call release(modesColumnH(1))
    call release(modesColumnH(2))
    call release(modesColumnH(3))
    call release(modesColumnH(4))
    call release(modesColumnH(5))
    call release(modesTableH)

! Add the quality extension to the image file.

    qualArrayDim(1) = nx
    qualArrayDim(2) = ny

    qualArrayH = addArray(outset, "QUALITY", Integer8, qualArrayDim)
    qualArrayImage => int8array2Data(qualArrayH)
    qualArrayImage=0

! Release handles on data.
    call release(qualArrayH)
!
!------------------------------------------------------------------------
!
    image = 0.0

    call random_number(background)
    background = background * 1000.0
    minFlux = nSigma * sqrt(background)
    maxFlux = minFlux * 1000.0  ! Was 100
    call random_number(xSource)
    call random_number(ySource)
    xSource = 5.0 + xSource * real(nx - 5)
    ySource = 5.0 + ySource * real(ny - 5)
    call random_number(flux)
    flux = flux * (maxFlux - minFlux) + minFlux
    call random_number(pa)
    pa = pa * 2.0 * pi
    do i = 1, nSource
        call random_number(r1)
        if (r1 > fracExtend) then
            sigX(i) = 0.5
            sigY(i) = 0.5
        else
            sigX(i) = 2.0
            call random_number(sigY(i))
            sigY(i) = (0.5 + sigY(i)) * sigX(i)
        end if
        sigXTimesSigY = sigX(i) * sigY(i)
        sigX(i) = 2.0 * sigX(i) ** 2
        sigY(i) = 2.0 * sigY(i) ** 2
    end do

    rSourceFlux = 0.0
    do i = 0, nx - 1
        do j = 0, ny - 1
            x = real(i) + 0.5
            y = real(j) + 0.5
         
            tot = 0.0
            do k = 1, nSource
                r = sqrt((x - xSource(k)) ** 2 + (y - ySource(k)) ** 2)
                if (r .lt. 20.0) then

                    delta = atan2(y - ySource(k), x - xSource(k))

                    xDash = r * cos(delta - pa(k))
                    yDash = r * sin(delta - pa(k))

                    ex1 = -(xDash ** 2 / sigX(k))

                    ex2 = -(yDash ** 2 / sigY(k))


                    f =  0.159154943 / sigXTimesSigY(k)

!--- VNY 2.11.2001
                    if((ex1+ex2) .lt. -10) then
                       f=0.0
                    else              
                       f = f * flux(k) * exp(ex1 + ex2)
                    end if
!--- 
                    rSourceFlux(k) = rSourceFlux(k) + f
                    tot = tot + f
                end if
            end do
            image(i, j) = image(i, j) + tot
        end do
    end do

    do i = 0, nx - 1
        do j = 0, ny - 1
            f = gasdev()
            meanval = image(i, j) + background
            image(i, j) = meanval + f * sqrt(meanval) ! gaussian deviate
        end do
    end do

    call setAttribute(outSet, "EXPOSURE", exposure)
    call setAttribute(outset, "BLANK", 999999.9)   ! Added by CB
    call setAttribute(outSet, "BINAX1", 1)
    call setAttribute(outSet, "BINAX2", 1)
    call setAttribute(outSet, "BINBPE", .false.)
    call setAttribute(outSet, "FILTER", "V")
    call setAttribute(outSet, "DATAMODE", trim(dataMode))
    call release(outArray)
    call release(outSet)

    open(20, file = srcListFileName, status= "unknown", form="formatted")
   
    write(20, *) nSource
    do i = 1, nSource
        write(20,*) xSource(i), ySource(i), rSourceFlux(i)/exposure, &
        sigX(i), sigY(i), pa(i)
    end do
    close(unit=20)

end subroutine CreateTestImage

real function gasdev() result(output)

  integer, save :: iset = 0
  real :: fac, rsq, v1, v2, ranNumber
  real, save :: gset

  if(iset .eq. 0)then
     loop1: do
!---Test VNY 2.11.2001
        ranNumber=0.0

        call random_number(ranNumber)
        v1 = 2.0 * ranNumber - 1.0

        call random_number(ranNumber)
        v2 = 2.0 * ranNumber - 1.0
        rsq = v1 ** 2 + v2 ** 2
        if(rsq .ge. 1.0 .or. rsq .lt. 1e-6) then
           continue
        else
           exit loop1
        endif
     end do loop1
     fac = sqrt(-2.0 * log(rsq) / rsq)
     gset = v1 * fac
     output = v2 * fac
     iset=1
  else
     output = gset
     iset = 0
  endif
  return
end function gasdev















